/* lpg.c - routines for bitmap graphics */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "lpg.h"
#include "prsetup.h"

#define MXNO    512

OVRLAY *getovrly();
int     lpg_x();
int     lpg_y();
void    lpgout();
void    lpgovrl();

extern int debug;
extern int network;
extern int printer;
extern FILE *lgf;

OVRLAY *ovrly[MXNO] = {NULL};
double   a_coef = 1, b_coef = 0, c_coef = 0;
double   dotdiam = 0.0;
int     adupi_x = 0;
int     adupi_y = 0;
int     bpp = 1;
int     novrly = 0;
int     pglen = 0;
int     pgwid = 0;
unsigned char mask[BPS] = {0};

void    (*pr_idle)();
void    (*pr_page)();
void    (*pr_out)();
void    (*pr_end)();
void    (*pr_rule)();
void    (*pr_line)();
void    (*pr_gray)();

static float *gray_map;
static int dqo = 0;
static int bc = 0, ibcnt = 0;
static time_t thistime = 0;
static time_t lasttime = 0;
static long pgc = 0, pgn = 1;
static unsigned char *buf = NULL;
static unsigned char *py = NULL; /* cumulative dot probability */

/*
 * lpgini - initialize line printer graphics
 *      ipmax - max number of dot positions across page
 */

void
lpgini()
{
    int     i, d;

    if (printer < 0 || printer >= NDEV) {
	fprintf(lgf, "Invalid printer selection: %d\n", printer);
	exit(1);
    }
    pr_setup[printer]();

    ibcnt = pgwid * bpp;
    pgn = 1;
    if (buf != NULL)
	free (buf);
    buf = (unsigned char *) calloc((size_t) ibcnt, 1);
    if (py != NULL)
	free(py);
    py = (unsigned char *) calloc((size_t) pgwid, 1);
    if (mask[0] == 0)
	for (i = 0; i < BPS; i++)
	    mask[i] = 1 << (BPS - 1 - i);

    novrly = 0;
    dqo = 0;
    pgc = 0;
    bc = 0;
    d = (bpp > 0) ? 256 >> bpp : 256;
    for (i = 0; i < pgwid; i++)
	py[i] = rand() % d;
}

void
lpgidle()
{
#ifdef MSDOS
    /*
       The purpose of this routine is to make certain that there is never
       more than few few seconds before something gets output to the printer
       port, so that shared printers are not reallocated to another PC.
    */
    if (network) {
	if (time(&thistime) > lasttime) {	/* idle every second */
	    pr_idle();
	    lasttime = thistime;
	}
    }
#endif
}

int
lpg_x(x)
double  x;
{
    return ((int) (x * adupi_x));
}

int
lpg_y(y)
double  y;
{
    return ((int) (y * adupi_y));
}

double
lpg_d()
{
    return (dotdiam);
}

int
lpg_p()
{
    pgn = 1;
    return (pgn);
}

void
lpggrmp(float *mp)
{
    gray_map = mp;  // map from index of px to range 0-1
}

/* lpggram - define gray level for one-wide strip of dots, with freq. map
 *      px - gray level array (0-255)
 *      nx - number of elements in px
 *      sp - starting position
 *      np - number of positions
 *      bo - bit offset within current output strip
 */
void
lpggram(unsigned char *px, int nx, int sp, int np, int bo)
{
    double  a, b, c, d, pl;
    float  *mp;
    int     ip, ii, jp, jj, ia, ib, m, p, pp, nint, ofst;

    mp = gray_map;  // map from index of px to range 0-1
    if ((sp + np) > pgwid)
	np = pgwid - sp;
    d = 1.0 / (np - 1);
    if (bpp == 2) {
	nint = 3;
	ofst = 0;
    } else if (bpp == 4) {
	nint = 12;
	ofst = 256 / nint;
    } else if (bpp == 8) {
	nint = 254;
	ofst = 0;
    } else {
	nint = 1;
	ofst = 0;
    }
    for (ip = 0, ib = 1; ip < np; ip++) {
	jp = sp + ip;
	pl = ip * d;
	while (ib < nx && mp[ib] < pl)
	    ib++;
	ia = ib - 1;
	if (ib >= nx) {
	    p = px[ia] + py[jp];
	} else {
	    c = mp[ib] - mp[ia];
	    a = (c <= 0) ? 0 : (mp[ib] - pl) / c;
	    b = 1 - a;
	    p = (int) (a * px[ia] + b * px[ib] + py[jp]);
	}
	m = mask[bo];
	if (bpp == 1) {
	    if (p >= 256) {
		buf[jp] |= m;
		p -= 256;
	    }
	} else if (bpp == 8) {
	    buf[jp * 8 + bo] = 1 + (p * nint + 128) / 256;
	    p = 0;
	} else {
	    jj = jp * bpp;
	    p += ofst;
	    for (ii = 0; ii < bpp; ii++) {
		pp = (256 * (1 << (bpp - 1 - ii))) / nint;
		if (p >= pp) {
		    buf[jj + ii] |= m;
		    p -= pp;
		}
	    }
	}
	py[jp] = p;
    }
}

/* lpgadv - advance the page n dot widths */

void
lpgadv(int n)
{
    while (n-- > 0) {
	bc++;			/* increment the byte bit counter */
	pgc++;			/* increment the page dot counter */
	if (bc >= BPS) {
	    lpgout();
	}
    }
}

/* lpgadvto - advance the page to position n */

void
lpgadvto(int n)
{
    while (n > pgc) {
	bc++;			/* increment the byte bit counter */
	pgc++;			/* increment the page dot counter */
	if (bc >= BPS) {
	    lpgout();
	}
    }
}

/* lpgpage - advance to top of next page */

void
lpgpage()
{
    int i, d;
    OVRLAY *b;

    if (pgc > pglen) {
	pgc %= pglen;
	pgn++;
    }
    if (pgc > 0) {
	while (pgc < pglen) {
	    bc++;		/* increment the byte bit counter */
	    pgc++;		/* increment the page dot counter */
	    if (bc >= BPS) {
		lpgout();
	    }
	}
	pgc = 0;
	pgn++;
    }
    pr_page();

    for (i = 0; i < novrly; i++) {
	b = ovrly[i];
	if (b->pbb != NULL) {
	    free(b->pbb);
	    b->pbb = NULL;
	}
    }
    novrly = 0;
    dqo = 0;
    pgc = 0;
    bc = 0;
    d = (bpp > 0) ? 256 >> bpp : 256;
    for (i = 0; i < pgwid; i++)
	py[i] = rand() % d;
}

/*
 * lpgout - flush buf out to line printer
 */
void
lpgout()
{
    int     i, n;

    if (pgc > pglen) {
	pgc %= pglen;
	pgn++;
    }
    if (novrly > 0)
	lpgovrl();

/* strip trailing zeros */
    for (n = ibcnt; n > 0 && buf[n - 1] == '\0'; n--)
	continue;
    if (bpp > 1 && (n % bpp) != 0)
	n += (bpp - (n % bpp));

/* send line to printer */
    if (pr_out)
	pr_out(buf, n);

/* zero buffer */
    for (i = 0; i < n; i++)
	buf[i] = 0;
    bc = 0;
}

/* lpgend - flush the buffer and exit graphics mode
 */
void
lpgend()
{
    if (pgc > 0) {
	dqo++;			/* dequeue overlays */
	while (bc > 0 || novrly > 0) {
	    bc++;		/* increment the byte bit counter */
	    pgc++;		/* increment the page dot counter */
	    if (bc >= BPS) {
		lpgout();
	    }
	}
    }
    pr_end(1);
}

/* lpgstop - quick exit from graphics mode
 */
void
lpgstop()
{
    pr_end(0);
}

void
lpgovrl()
{
    double  w;
    long    i, np, *xx, *yy;
    int     o, i0, i1, ii, j, j0, j1, j2, jj, m, n;
    OVRLAY *ov;

    // overlay gray strips first
    for (o = 0; o < novrly; o++) {
	if ((ov = ovrly[o]) == NULL) {
	    continue;
	}
	i = pgc - ov->xllc - BPS;
	if (i > ov->xlen && !ov->outs) {	/* passed overlay ? */
	    if (ov->otyp == OV_GRAY && pr_gray) {
		pr_gray(ov->xllc, ov->yllc, ov->xlen, ov->ylen, 
		    ov->xbyt, ov->pbb, gray_map);
	    }
	}
    }
    for (o = 0; o < novrly; o++) {
	if ((ov = ovrly[o]) == NULL) {
	    fprintf(lgf, "Overlay %d has null pointer\n", o);
	    continue;
	}
	i = pgc - ov->xllc - BPS;
	if (i > ov->xlen && !ov->outs) {	/* passed overlay ? */
	    if (ov->otyp == OV_RULE && pr_rule) {
		pr_rule(ov->xllc, ov->yllc, ov->xlen, ov->ylen);
	    } else if (ov->otyp == OV_LINE && pr_line) {
		n = ov->yllc;
		w = 0.001 * ov->ylen;
		xx = (long *) ov->pbb;
		yy = xx + n;
		pr_line(n, xx, yy, w);
	    }
            ov->outs = 1;
	}
	if (i <= -(BPS) || i >= ov->xlen) {	/* overlay out of range ? */
	    if (ov->outs || dqo) {
		if (ov->pbb != NULL) {
		    free(ov->pbb);
		    ov->pbb = NULL;
		}
		novrly--;
		if (o < novrly) {
		    ovrly[o--] = ovrly[novrly];
		    ovrly[novrly] = ov;
		}
	    }
	} else if (ov->otyp == OV_GRAY) {		/* gray ? */
	    if (!pr_gray) {
		for (ii = 0;  ii < ov->xlen; ii++) {
		    lpggram(ov->pbb, ov->xbyt, ov->yllc, ov->ylen, ov->xoff + ii);
		}
	    }
	} else if (ov->otyp == OV_RULE) {		/* rule ? */
	    if (!pr_rule) {
		np = ov->xlen;
		i0 = (i < 0) ? -i : 0;
		i1 = (np - i < BPS) ? np - i : BPS;
		m = 0;
		for (ii = i0; ii < i1; ii++)
		    m |= mask[ii];
		n = ov->ylen;
		j0 = ov->yllc;
		j1 = limit(-j0, 0, pgwid - j0);
		j2 = limit(-j0, n, pgwid - j0);
		if (bpp == 1) {
		    for (j = j1; j < j2; j++)
			buf[j + j0] |= m;
		} else if (bpp == 8) {
		    for (j = j1; j < j2; j++) {
			jj = (j + j0) * bpp;
		        for (ii = 0; ii < bpp; ii++) {
			    if (m & (1 << (bpp - 1 - ii))) {
			         buf[jj + ii] = 255;
			    }
		        }
		    }
		} else {
		    for (j = j1; j < j2; j++) {
			jj = (j + j0) * bpp;
			for (ii = 0; ii < bpp; ii++)
			    buf[jj + ii] |= m;
		    }
		}
		ov->outs = 1;
	    }
	} else if (ov->otyp == OV_BTBK) {		/* bit block ? */
	    i0 = i / BPS;
	    n = ov->ylen;
	    j0 = ov->yllc;
            j1 = limit(-j0, 0, pgwid - j0);
            j2 = limit(-j0, n, pgwid - j0);
	    if (bpp == 1) {
		for (j = j1; j < j2; j++)
		    buf[j + j0] |= ov->pbb[j * ov->xbyt + i0];
	    } else if (bpp == 8) {
		for (j = j1; j < j2; j++) {
		    jj = (j + j0) * bpp;
		    m = ov->pbb[j * ov->xbyt + i0];
		    for (ii = 0; ii < bpp; ii++) {
			if (m & (1 << (bpp - 1 - ii))) {
			     buf[jj + ii] = 255;
			}
		    }
		}
	    } else {
		for (j = j1; j < j2; j++) {
		    jj = (j + j0) * bpp;
		    m = ov->pbb[j * ov->xbyt + i0];
		    for (ii = 0; ii < bpp; ii++)
			buf[jj + ii] |= m;
		}
	    }
	    ov->outs = 1;
	}
    }
}

/* lpgrule - define a filled rectangle for line printer graphics */

void
lpgrule(long xllc, long yllc, long xlen, long ylen)
{
    OVRLAY *ov;

    if (xlen <= 0.0 || ylen <= 0.0)
	return;
    if ((ov = getovrly()) == NULL)
	return;
    ov->xllc = xllc;
    ov->yllc = yllc;
    ov->xlen = xlen;
    ov->ylen = ylen;
    ov->xoff = 0;
    ov->yoff = 0;
    ov->xbyt = 0;
    ov->outs = 0;
    ov->otyp = OV_RULE;
    ov->pbb = NULL;
}

/* lpggray - define a gray-strip for line printer graphics */

void
lpggray(unsigned char *px, int nx, int yllc, int ylen, int xlen)
{
    OVRLAY *ov;

    if (xlen <= 0.0 || ylen <= 0.0)
	return;
    if ((ov = getovrly()) == NULL)
	return;
    ov->xllc = pgc;
    ov->yllc = yllc;
    ov->xlen = xlen;
    ov->ylen = ylen;
    ov->xoff = bc;
    ov->yoff = 0;
    ov->xbyt = nx;
    ov->outs = 0;
    ov->otyp = OV_GRAY;
    ov->pbb = (char *) malloc(nx + 1);
    memcpy(ov->pbb, px, nx + 1);
}

OVRLAY *
getovrly()
{
    OVRLAY *ov;

    if (mask[0] == 0) {
	fprintf(lgf, "Bit mask initialization error\n");
	return (NULL);
    }
    if (novrly >= MXNO) {
	fprintf(lgf, "Overlay limit exceeded: %d\n", novrly);
	return (NULL);
    }
    if (ovrly[novrly] == NULL)
	ovrly[novrly] = (OVRLAY *) calloc(sizeof(OVRLAY), 1);
    ov = ovrly[novrly];
    if (ov == NULL) {
	fprintf(lgf, "Overlay pointer initialization error\n");
	return (NULL);
    }
    ov->otyp = OV_BTBK;
    ov->pbb = NULL;
    novrly++;
    return (ov);
}

/* lpgcomp - compenstate for printer imaging characteristics */

double
lpgcomp(a)
double  a;
{
    return ((a_coef + (b_coef + c_coef * a) * a) * a);
}
