/* specto - high resolution spectrographs on line printer */

#define MAIN
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <fcntl.h>
#include <signal.h>
#include "specto.h"
#include "frame.h"
#include "cparam.h"
#include "prinfo.h"
#include "version.h"

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#define _fileno fileno
#define _access access
#endif /* WIN32 */

#define MF  16			/* maximum number of file names */
#define MP  256			/* maximum number of plot points */
#ifndef PI
#define PI  3.1415926535898
#endif

int     guess_file_type();
int     lpg_x();
int     lpg_y();
int     lpg_p();
int     match();
void    chkenv();
void    defflg();
void    defpar();
void    doplot();
void    dotext();
void    expndstr();
void    getpar();
void    spckill();
void    leave();
void    lpgpage();
void    lpgend();
void    lpgstop();
void    lpgdraw();
void    lpgtext();
void    lpgchup();
void    lpgchlw();
void    lpgchsz();
void    lpgchal();
void    lpgchxt();
void    read_args();
void    select_printer();
void    set_binary();
void    set_par();
void    trim();
void    process();
void    finish_page();
void    setup_lpg();

struct _flags flags_;
struct _bpar bpar_;
struct _fpar fpar_;
struct _ucon ucon_;
struct _wave wave_;
struct _tick tick_;
struct _grid grid_;

char   *bparnam[] = {
    "_bpar",
    "aronum", "aswv", "bits", "date", "eaxis", "elabel", "fannot", "faxis",
    "flabel", "ftpi", "ftpw", "grid", "hfemph", "invwav",
    "mfft", "mhal", "mval", "mxtnt", "newpage", "nflim", "nlpc", "over",
    "periodic", "prenrg", "prspec", "prwave", "rmdc", "sign", "swap",
    "tannot", "taxis", "tlabel", "toba", 
    "waxis", "window", "wlabel", "wrfile",
};
char   *fparnam[] = {
    "_fpar",
    "agcsp", "annlwt", "annsiz", "arosiz", "arotip", "arowid", "axlnwd", 
    "bandwidth", "contrast", "darkness", "eeps", "elen", "ellc", "emax", 
    "emin", "enrm", "fgrid", "fint", "floor", "fmax", "fmin", "foct", 
    "fpma", "fpmb", "fpmc",  "head", "lablwt", "labsiz", "msgang", "msglwt", 
    "msgsiz", "pltlwt", "rate", "tail", "ticsiz", "tmax", "tmin", "token", 
    "tpage",  "windwidth", "wlen", "wllc", "xlen", "xllc", "xpage", "ylen", 
    "yllc", "ypage"
};
char    file_name[256];
char    pfn[256];
char    time_date[40];
int     nbpar = sizeof(bparnam) / sizeof(bparnam[0]);
int     nfpar = sizeof(fparnam) / sizeof(fparnam[0]);
int     printer = 0;
int     network = 0;
int     noheader = 0;
int     noquery = 0;
int     p_gmode = 0;
int     p_color = 0;
int     setup_lpg_called = 0;
int     inf_type = 0;
int     parft = 0;
int     sd_asc = 0;
int     sdf_type = 0;
int     cpu_lsb = 0;
FILE   *prn = NULL;
FILE   *lgf = NULL;

static void setlsb();
static void usage();

static float deg2rad = (float) (PI / 180);
static char clfn[MF][40];
static char template[40] = "template.spc";
static int nclfn = 0;
static int dfp = 0;
static double x[MP], y[MP];

void
newext(f, e)
char   *f, *e;
{
    char   *s;

    s = f;
    while (*s != '\0')		/* locate end of filename */
	s++;
    while (*s != '\\' && *s != '/' && s > f)	/* backup to start of name */
	s--;
    while (*s != '\0' && *s != '.')	/* locate end of name */
	s++;
    while (*e != '\0')		/* copy extension */
	*s++ = *e++;
    *s = '\0';
}

void
defflg()
{
    printer = dfp;
    flags_.debug = 0;
    flags_.verbose = 1;
}

void
chkenv()
{
    char    eb[256];
    char   *eav[32], *evs;
    int     eac;
    static char *evn = "SPECTO";

    if ((evs = getenv(evn)) == NULL)
	return;
    fprintf(lgf, "%s=%s\n", evn, evs);
    strcpy(eb, evs);
    eav[0] = evn;
    eac = 1;
    for (evs = eb; *evs != '\0'; evs++) {
	while (*evs == ' ' || *evs == '\t')	/* skip blanks */
	    evs++;
	if (*evs == '\0')	/* quit if EOS */
	    break;
	eav[eac++] = evs;	/* point to arg */
	while (*evs != ' ' && *evs != '\t' && *evs != '\0')	/* skip arg */
	    evs++;
	if (*evs == '\0')	/* quit if EOS */
	    break;
	*evs = '\0';		/* terminate arg */
    }
    if (eac > 1)
	read_args(eac, eav);
}

void
read_args(ac, av)
int     ac;
char  **av;
{
    nclfn = 0;
    av++;
    ac--;
    while (ac > 0) {
	if (av[0][0] != '-') {
	    if (strchr(av[0], '=') != NULL) {	/* parameter */
		set_par(av[0]);
	    } else if (nclfn < MF) {	/* file name */
		strcpy(clfn[nclfn], av[0]);
		trim(clfn[nclfn]);	/* ? */
		nclfn++;
	    }
	} else {		/* process command line options */
	    switch (av[0][1]) {
	    case 'a':
		sd_asc++;
		break;
	    case 'c':
		p_color++;
		break;
	    case 'd':
		flags_.debug++;
		break;
	    case 'h':
		if (ac > 1 && atoi(av[1]) > 0)
		    fpar_.head = (float) atof(av[1]);
		av++;
		ac--;
		break;
	    case 'n':
                if (av[0][2] == 'h')
                    noheader++;
                else if (av[0][2] == 'q')
                    noquery++;
                else
                    network++;
		break;
	    case 'o':
		if (prn != NULL)
		    fclose(prn);
		if (ac > 1 && (prn = fopen(av[1], "wb")) == NULL) {
		    fprintf(lgf, "can't open output file %s\n", av[1]);
		    exit(0);
		}
		av++;
		ac--;
		break;
	    case 'p':
	    case 'P':
		select_printer(av[0] + 2);
		break;
	    case 'q':
		flags_.verbose = 0;
		break;
	    case 'r':
		if (ac > 1 && atoi(av[1]) > 0)
		    fpar_.rate = (float) atof(av[1]);
		av++;
		ac--;
		break;
	    case 't':
		if (ac > 1 && atoi(av[1]) > 0)
		    fpar_.tail = (float) atof(av[1]);
		av++;
		ac--;
		break;
	    }
	}
	av++;
	ac--;
    }
}

void
defpar()
{
    bpar_.aronum = 0;
    bpar_.aswv = 1;
    bpar_.bits = 0;
    bpar_.date = 0;
    bpar_.eaxis = 1;
    bpar_.elabel = 1;
    bpar_.fannot = 1;
    bpar_.faxis = 1;
    bpar_.flabel = 1;
    bpar_.ftpi = 64;		/* max. Fourier transforms per inch */
    bpar_.ftpw = 64;		/* max. Fourier transforms per window */
    bpar_.grid = 1;
    bpar_.hfemph = 1;
    bpar_.invwav = 0;
    bpar_.mfft = 0;
    bpar_.mhal = 1;
    bpar_.mval = 4;
    bpar_.mxtnt = 0;
    bpar_.newpage = 1;
    bpar_.nflim = 1;		/* Nyquist frequency limit ? */
    bpar_.nlpc = 0;
    bpar_.over = 0;
    bpar_.periodic = 0;
    bpar_.prenrg = 1;
    bpar_.prspec = 1;
    bpar_.prwave = 1;
    bpar_.rmdc = 0;
    bpar_.sign = 0;
    bpar_.swap = 0;
    bpar_.tannot = 1;
    bpar_.taxis = 1;
    bpar_.tlabel = 1;
    bpar_.toba = 1;
    bpar_.waxis = 1;
    bpar_.window = 2;
    bpar_.wlabel = 1;
    bpar_.wrfile = 0;

    fpar_.agcsp = 0.5;
    fpar_.annlwt = 1.0;
    fpar_.annsiz = (float) 0.08;
    fpar_.arosiz = (float) 0.15;
    fpar_.arotip = (float) 0.20;
    fpar_.arowid = (float) 0.75;
    fpar_.axlnwd = (float) 0.01;
    fpar_.bandwidth = 400.0;
    fpar_.contrast = 0.50;	/* contrast of spectrograph */
    fpar_.darkness = 0.50;	/* darkness of spectrograph */
    fpar_.elen = 0.25;
    fpar_.ellc = 5.75;
    fpar_.emax = 0.0;
    fpar_.emin = -40.0;
    fpar_.enrm = (float) 0.1;
    fpar_.eeps = 200.0;
    fpar_.fmax = 8000.0;
    fpar_.fmin = 0.0;
    fpar_.fgrid = 1000.0;	/* grid spacing */
    fpar_.fint = 1000.0;	/* annotation spacing */
    fpar_.floor = 100;		/* intensity transformation exponent */
    fpar_.foct = 0;		/* number of octaves on frequency scale */
    fpar_.fpma = 0;		/* frequency-to-place parameter a */
    fpar_.fpmb = 0;		/* frequency-to-place parameter b */
    fpar_.fpmc = 0;		/* frequency-to-place parameter c */
    fpar_.head = -1.0;		/* sdf header size (bytes) */
    fpar_.lablwt = 1.0;
    fpar_.labsiz = (float) 0.1;
    fpar_.msgang = 0.0;
    fpar_.msglwt = 1.0;
    fpar_.msgsiz = (float) 0.1;
    fpar_.pltlwt = 1.0;
    fpar_.rate = 0.0;		/* sdf sampling rate (samp/sec) */
    fpar_.tail = 0.0;		/* sdf trailer size (bytes) */
    fpar_.ticsiz = (float) 0.08;
    fpar_.tmax = 0.0;
    fpar_.tmin = 0.0;
    fpar_.token = 1.0;
    fpar_.tpage = 0.0;
    fpar_.windwidth = 0;
    fpar_.wlen = 1.0;
    fpar_.wllc = 6.25;
    fpar_.xlen = 8.5;
    fpar_.xllc = 1.25;
    fpar_.xpage = 11.0;
    fpar_.ylen = 4.5;
    fpar_.yllc = 1.0;
    fpar_.ypage = 8.0;
#ifdef JBA_DEFAULTS
    fpar_.agcsp = 0;		/* 0.5; */
    fpar_.elen = 1;		/* 0.25; */
    fpar_.ellc = 5.25;		/* 5.75; */
    fpar_.emin = -70;		/*-40.0;*/
    fpar_.wllc = 6.5;		/* 6.25; */
    fpar_.yllc = .5;		/* 1.0; */
#endif
}

void
getpar(fn)
char   *fn;
{
    FILE   *pfp;
    char    str[256], *val, *bpar, *s;
    float  *fpar;
    int     i;
    time_t  clock_time;
    static char *ft_str[] = {
	"SPC", "SDF", "TSD", "ILS", "SND", "LST", "MAT", "NOH", "WAV", "TOK"
    };

    (void) time(&clock_time);
    strcpy(time_date, ctime(&clock_time));
    strcpy(file_name, fn);
    trim(file_name);
    strcpy(pfn, fn);
    if (_access(pfn, 0) != 0) {
	newext(pfn, ".spc");
	if (_access(pfn, 0) != 0) {
	    (void) fprintf(lgf, "can't open file %s\n", fn);
	    return;
	}
    }
    inf_type = parft = guess_file_type(pfn, sd_asc);
    fprintf(lgf, "%s file: %s\n", ft_str[sdf_type], pfn);
    if (inf_type == FT_SDF || inf_type == FT_TSD) {
        if (nclfn == 1 && _access(template, 0) == 0) {	/* template? */
	    nclfn=2;
	    strcpy(clfn[1], pfn);
	    strcpy(pfn, template);
	    inf_type = FT_SPC;
	}
    }
    if (inf_type == FT_SDF || inf_type == FT_TSD) {
	strcpy(wave_.sdfn, pfn);
	process(inf_type);
    } else {			/* FT_SPC */
	if (strchr(pfn, '.') == NULL)
	    strcat(pfn, ".spc");/* add default extension */
	pfp = fopen(pfn, "r");
	if (pfp == NULL) {
	    (void) fprintf(lgf, "can't open parameter file %s\n", pfn);
	    return;
	}
	while (fgets(str, 256, pfp) != NULL) {
	    expndstr(str, 256);
	    i = strlen(str);
	    str[i - 1] = '\0';	/* remove <nl> from end */
	    for (s = str; *s == ' '; s++)	/* remove leading blanks */
		continue;
	    if (*s == ';') {	/* comment */
		continue;
	    } else if (strchr(s, '"') != NULL) {	/* text */
		dotext(s);
	    } else if (strchr(s, '=') != NULL) {	/* parameter */
		val = strchr(s, '=') + 1;
		bpar = &bpar_._bpar;
		for (i = 0; i < nbpar; i++)
		    if (match(s, bparnam[i]))
			bpar[i] = atoi(val);
		fpar = (float *) &fpar_;
		for (i = 0; i < nfpar; i++)
		    if (match(s, fparnam[i]))
			fpar[i] = (float) atof(val);
	    } else if (match(s, "ilsfile") || match(s, "sdfile")
		|| match(s, "tsdfile") || match(s, "file")) {
		while (*s != ' ' && *s != '\0')
		    s++;
		while (*s == ' ')
		    s++;
		strcpy(wave_.sdfn, s);
		trim(wave_.sdfn);
		while (i-- > 0 && wave_.sdfn[i] == ' ')
		    wave_.sdfn[i] = 0;
		inf_type = guess_file_type(wave_.sdfn, 1);
		process(inf_type);
	    } else if (match(s, "plot")) {
		doplot(s);
	    }
	}
	(void) fclose(pfp);
	finish_page();
    }
}

void
leave(int status)
{
    fflush(lgf);
    if (setup_lpg_called) {
	if (status != 0) {
	    (void) signal(SIGINT, SIG_IGN);
	    lpgstop();
	} else {
	    if (bpar_.newpage)
		lpgpage();
	    lpgend();
	}
	setup_lpg_called = 0;
    }
    if (prn) {
        fclose(prn);
	prn = NULL;
    }
#ifndef WIN32
    fprintf(stderr, "\n%s  %s.\n", CPR, BTH, DST);
    exit(status);
#endif
}

int
match(s, p)
char   *s, *p;
{
    int     i;

    for (i = 0; i < 4; i++) {
	if (*s >= 'A' && *s <= 'Z')
	    *s += 'a' - 'A';
	if (*s++ != *p++)
	    break;
    }
    return (i == 4);
}

void
draw_arrow(double x1, double y1, double x2, double y2, double lw)
{
    int     n;
    double  w, h, r, t, dx, dy, xx, yy, x[5], y[5];

    if (fpar_.arosiz > 0 && fpar_.arowid > 0) {
        dx = x2 - x1;
	dy = y2 - y1;
	r = sqrt(dx * dx + dy * dy);
	if (r > 0) {
	    dx /= r;
            dy /= r;
	    h = fpar_.arosiz;
	    w = h * fpar_.arowid / 2;
            t = h * fpar_.arotip;
	    n = 0;
	    xx = x2;
	    yy = y2;
            x[n] = xx - t * dx;
	    y[n] = yy - t * dy;
	    n++;
	    x[n] = xx - h * dx - w * dy;
	    y[n] = yy - h * dy + w * dx;
	    n++;
	    x[n] = xx;
	    y[n] = yy;
	    n++;
	    x[n] = xx - h * dx + w * dy;
	    y[n] = yy - h * dy - w * dx;
	    n++;
	    x[n] = xx - t * dx;
	    y[n] = yy - t * dy;
	    n++;
	}
    }
    lpgdraw(n, x, y, lw);
}

void
doplot(char *s)
{
    double lw;
    int    i, n;

    setup_lpg();
    for (i = 0; i < MP && *s; i++) {
        while (*s != ' ' && *s != '\0')    /* skip non-blank  */
	    s++;
        while (*s == ' ')			/* skip blank */
	    s++;
	if (*s == '\0')
	    break;
        x[i] = atof(s);			/* get x value */
        while (*s != ' ' && *s != '\0')    /* skip non-blank  */
	    s++;
        while (*s == ' ')			/* skip blank */
	    s++;
	if (*s == '\0')
	    break;
        y[i] = atof(s);			/* get y value */
    }
    n = i;
    lw = fpar_.pltlwt * 0.02;
    lpgdraw(n, x, y, lw);
    if (n > 1) {
	if (bpar_.aronum == 1 || bpar_.aronum == 2)
	    draw_arrow(x[n-2], y[n-2], x[n-1], y[n-1], lw);
	if (bpar_.aronum == 2 || bpar_.aronum == -1)
	    draw_arrow(x[1], y[1], x[0], y[0], lw);
    }
}

void
dotext(s)
char   *s;
{
    char   *eq;
    double  x, y, a;

    if (*s == '"')
	return;
    x = atof(s);
    while (*s != ' ' && *s != '\0' && *s != '"')
	s++;
    while (*s == ' ' && *s != '"')
	s++;
    if (*s == '"')
	return;
    y = atof(s);
    while (*s != ' ' && *s != '\0' && *s != '"')
	s++;
    while (*s == ' ' && *s != '"')
	s++;
    if (*s != '"')
	return;
    eq = strchr(++s, '"');
    if (eq == NULL)
	return;
    *eq = '\0';
    setup_lpg();
    a = fpar_.msgang * deg2rad;
    lpgchup(-sin(a), cos(a));
    lpgchlw(fpar_.msglwt);
    lpgchsz(fpar_.msgsiz);
    lpgchal(bpar_.mhal, bpar_.mval);
    lpgchxt(bpar_.mxtnt);
    lpgtext(x, y, s);
}

void
select_printer(s)
char   *s;
{
    int     i, n;

    printer = 0;
    p_gmode = 0;
    for (i = 0; i < npr; i++) {
	n = strlen(pr_info[i].n);
	if (strncmp(s, pr_info[i].n, n) == 0) {
	    printer = pr_info[i].p;
	    p_gmode = pr_info[i].m;
	    break;
	}
    }
}

void
expndstr(str, n)
char   *str;
int     n;
{
    int     i;
    char   *s, *b, buf[256];
    time_t  clock_time;

    if (*str == '\0' || strchr(str, '%') == NULL)
	return;
    b = buf;
    for (s = str; *s != '\0'; s++) {
	if (*s != '%') {
	    *b++ = *s;
	} else {
	    s++;
	    if (*s == '%') {
		*b++ = '%';
	    } else if (*s >= '0' && *s <= '9') {
		i = *s - '0';
		if (i >= nclfn) {
		    fprintf(lgf, "WARNING: don't have arg #%d\n", i);
		} else {
		    strcpy(b, clfn[i]);
		    b = buf + strlen(buf);
		}
	    } else if (*s == 'd') {
		(void) time(&clock_time);
		strcpy(b, ctime(&clock_time));
		b = buf + strlen(buf) - 1;
		*b = '\0';
	    } else if (*s == 'p') {
		sprintf(b, "%d", lpg_p());
		b = buf + strlen(buf);
	    } else {
		*b++ = '%';
		*b++ = *s;
	    }
	}
    }
    *b = '\0';
    strcpy(str, buf);
}

void
trim(s)
char   *s;
{
    char   *p;

    for (p = s + strlen(s) - 1; *p == ' ' || *p == '\r' || *p == '\n'; p--)
	continue;
    p[1] = '\0';
}

void
set_par(s)
char   *s;
{
    char   *val, *bpar;
    float  *fpar;
    int     i;

    val = strchr(s, '=') + 1;
    for (i = 0; i < nbpar; i++)
	if (match(s, bparnam[i])) {
	    bpar = &bpar_._bpar;
	    bpar[i] = atoi(val);
	    break;
	}
    for (i = 0; i < nfpar; i++)
	if (match(s, fparnam[i])) {
	    fpar = &fpar_._fpar;
	    fpar[i] = (float) atof(val);
	    break;
	}
}

void
setlsb()
{
    short   a = 1;
    char   *b = (char *) (&a);

    cpu_lsb = b[1] ? 1 : 0;
}

void
usage()
{
    int     i;
    static char *help_text[] = {
	VER,
	"",
	"usage: specto [-options] [param=value] spc_file [arg1 arg2 ...]",
	"   or: specto [-options] [param=value] sd_file",
	"",
	"where 'spc_file' is a parameter file "
	"and 'sd_file' is a sampled-data file.",
	"",
	"options: ",
	"   -a      sampled-data file is ASCII",
	"   -c      use colors on EGA & VGA",
	"   -d      debug print",
	"   -h N    sdf header size is N bytes",
	"   -n      network printer connected to output",
	"   -nh     assume sdf has no header",
	"   -o F    output to file F",
	"   -Plj    output to LaserJet printer (150 dpi)",
	"   -Pps    output to Postscript printer (300x300 dpi)",
	"   -q      quiet mode",
	"   -r N    set default sample rate to N samples per second",
	"   -t N    sdf trailer size is N bytes",
    };
    static int n = sizeof(help_text) / sizeof(help_text[0]);

    for (i = 0; i < n; i++)
	fprintf(lgf, "%s\n", help_text[i]);
    exit(0);
}

int
specto(int ac, char **av)
{
    lgf = stderr;		/* error output to stderr */
    setlsb();			/* determine order of LSB */
    defflg();			/* set default options */
    defpar();			/* set default parameters */
    chkenv();
    read_args(ac, av);
    if (nclfn == 0)		/* must specify a file */
	usage();
    iniframe();		/* initialize frame */
    getpar(clfn[0]);		/* read parameters from first file */
    leave(0);
    return (0);
}

#ifndef WIN32

void
spckill()
{
    leave(1);
}

int
main(int ac, char **av)
{
#ifdef MSC
    (void) signal(SIGINT, (void (__cdecl *)(int))spckill);
    prn = _stdprn;		/* default output to stdprn */
#else /* MSC */
    (void) signal(SIGINT, spckill);
    prn = stdout;		/* default output to stdout */
#endif /* MSC */
    set_binary(_fileno(prn));	/* do binary output */
    (void) specto(ac, av);
    return (0);
}

#endif /* WIN32 */
